@@ -1,5 +1,6 @@
 # release_robot.py
 #
+# SPDX-License-Identifier: Apache-2.0 OR GPL-2.0-or-later
 # Dulwich is dual-licensed under the Apache License, Version 2.0 and the GNU
 # General Public License as public by the Free Software Foundation; version 2.0
 # or (at your option) any later version. You can redistribute it and/or
@@ -49,7 +50,7 @@ import re
 import sys
 import time
 
-from dulwich.repo import Repo
+from ..repo import Repo
 
 # CONSTANTS
 PROJDIR = "."
@@ -134,7 +135,7 @@ def get_current_version(projdir=PROJDIR, pattern=PATTERN, logger=None):
         current_version = matches.group(1)
     except (IndexError, AttributeError) as err:
         if logger:
-            logger.exception(err)
+            logger.debug("Pattern %r didn't match tag %r: %s", pattern, tag, err)
         return tag
     return current_version
 
