@@ -68,6 +68,180 @@ EXTENDED_FLAG_INTEND_TO_ADD = 0x2000
 
 DEFAULT_VERSION = 2
 
+# Index extension signatures
+TREE_EXTENSION = b"TREE"
+REUC_EXTENSION = b"REUC"
+UNTR_EXTENSION = b"UNTR"
+EOIE_EXTENSION = b"EOIE"
+IEOT_EXTENSION = b"IEOT"
+
+
+def _encode_varint(value: int) -> bytes:
+    """Encode an integer using variable-width encoding.
+
+    Same format as used for OFS_DELTA pack entries and index v4 path compression.
+    Uses 7 bits per byte, with the high bit indicating continuation.
+
+    Args:
+      value: Integer to encode
+    Returns:
+      Encoded bytes
+    """
+    if value == 0:
+        return b"\x00"
+
+    result = []
+    while value > 0:
+        byte = value & 0x7F  # Take lower 7 bits
+        value >>= 7
+        if value > 0:
+            byte |= 0x80  # Set continuation bit
+        result.append(byte)
+
+    return bytes(result)
+
+
+def _decode_varint(data: bytes, offset: int = 0) -> tuple[int, int]:
+    """Decode a variable-width encoded integer.
+
+    Args:
+      data: Bytes to decode from
+      offset: Starting offset in data
+    Returns:
+      tuple of (decoded_value, new_offset)
+    """
+    value = 0
+    shift = 0
+    pos = offset
+
+    while pos < len(data):
+        byte = data[pos]
+        pos += 1
+        value |= (byte & 0x7F) << shift
+        shift += 7
+        if not (byte & 0x80):  # No continuation bit
+            break
+
+    return value, pos
+
+
+def _compress_path(path: bytes, previous_path: bytes) -> bytes:
+    """Compress a path relative to the previous path for index version 4.
+
+    Args:
+      path: Path to compress
+      previous_path: Previous path for comparison
+    Returns:
+      Compressed path data (varint prefix_len + suffix)
+    """
+    # Find the common prefix length
+    common_len = 0
+    min_len = min(len(path), len(previous_path))
+
+    for i in range(min_len):
+        if path[i] == previous_path[i]:
+            common_len += 1
+        else:
+            break
+
+    # The number of bytes to remove from the end of previous_path
+    # to get the common prefix
+    remove_len = len(previous_path) - common_len
+
+    # The suffix to append
+    suffix = path[common_len:]
+
+    # Encode: varint(remove_len) + suffix + NUL
+    return _encode_varint(remove_len) + suffix + b"\x00"
+
+
+def _decompress_path(
+    data: bytes, offset: int, previous_path: bytes
+) -> tuple[bytes, int]:
+    """Decompress a path from index version 4 compressed format.
+
+    Args:
+      data: Raw data containing compressed path
+      offset: Starting offset in data
+      previous_path: Previous path for decompression
+    Returns:
+      tuple of (decompressed_path, new_offset)
+    """
+    # Decode the number of bytes to remove from previous path
+    remove_len, new_offset = _decode_varint(data, offset)
+
+    # Find the NUL terminator for the suffix
+    suffix_start = new_offset
+    suffix_end = suffix_start
+    while suffix_end < len(data) and data[suffix_end] != 0:
+        suffix_end += 1
+
+    if suffix_end >= len(data):
+        raise ValueError("Unterminated path suffix in compressed entry")
+
+    suffix = data[suffix_start:suffix_end]
+    new_offset = suffix_end + 1  # Skip the NUL terminator
+
+    # Reconstruct the path
+    if remove_len > len(previous_path):
+        raise ValueError(
+            f"Invalid path compression: trying to remove {remove_len} bytes from {len(previous_path)}-byte path"
+        )
+
+    prefix = previous_path[:-remove_len] if remove_len > 0 else previous_path
+    path = prefix + suffix
+
+    return path, new_offset
+
+
+def _decompress_path_from_stream(f, previous_path: bytes) -> tuple[bytes, int]:
+    """Decompress a path from index version 4 compressed format, reading from stream.
+
+    Args:
+      f: File-like object to read from
+      previous_path: Previous path for decompression
+    Returns:
+      tuple of (decompressed_path, bytes_consumed)
+    """
+    # Decode the varint for remove_len by reading byte by byte
+    remove_len = 0
+    shift = 0
+    bytes_consumed = 0
+
+    while True:
+        byte_data = f.read(1)
+        if not byte_data:
+            raise ValueError("Unexpected end of file while reading varint")
+        byte = byte_data[0]
+        bytes_consumed += 1
+        remove_len |= (byte & 0x7F) << shift
+        shift += 7
+        if not (byte & 0x80):  # No continuation bit
+            break
+
+    # Read the suffix until NUL terminator
+    suffix = b""
+    while True:
+        byte_data = f.read(1)
+        if not byte_data:
+            raise ValueError("Unexpected end of file while reading path suffix")
+        byte = byte_data[0]
+        bytes_consumed += 1
+        if byte == 0:  # NUL terminator
+            break
+        suffix += bytes([byte])
+
+    # Reconstruct the path
+    if remove_len > len(previous_path):
+        raise ValueError(
+            f"Invalid path compression: trying to remove {remove_len} bytes from {len(previous_path)}-byte path"
+        )
+
+    prefix = previous_path[:-remove_len] if remove_len > 0 else previous_path
+    path = prefix + suffix
+
+    return path, bytes_consumed
+
 
 class Stage(Enum):
     NORMAL = 0
@@ -95,6 +269,83 @@ class SerializedIndexEntry:
         return Stage((self.flags & FLAG_STAGEMASK) >> FLAG_STAGESHIFT)
 
 
+@dataclass
+class IndexExtension:
+    """Base class for index extensions."""
+
+    signature: bytes
+    data: bytes
+
+    @classmethod
+    def from_raw(cls, signature: bytes, data: bytes) -> "IndexExtension":
+        """Create an extension from raw data.
+
+        Args:
+          signature: 4-byte extension signature
+          data: Extension data
+        Returns:
+          Parsed extension object
+        """
+        if signature == TREE_EXTENSION:
+            return TreeExtension.from_bytes(data)
+        elif signature == REUC_EXTENSION:
+            return ResolveUndoExtension.from_bytes(data)
+        elif signature == UNTR_EXTENSION:
+            return UntrackedExtension.from_bytes(data)
+        else:
+            # Unknown extension - just store raw data
+            return cls(signature, data)
+
+    def to_bytes(self) -> bytes:
+        """Serialize extension to bytes."""
+        return self.data
+
+
+class TreeExtension(IndexExtension):
+    """Tree cache extension."""
+
+    def __init__(self, entries: list[tuple[bytes, bytes, int]]) -> None:
+        self.entries = entries
+        super().__init__(TREE_EXTENSION, b"")
+
+    @classmethod
+    def from_bytes(cls, data: bytes) -> "TreeExtension":
+        # TODO: Implement tree cache parsing
+        return cls([])
+
+    def to_bytes(self) -> bytes:
+        # TODO: Implement tree cache serialization
+        return b""
+
+
+class ResolveUndoExtension(IndexExtension):
+    """Resolve undo extension for recording merge conflicts."""
+
+    def __init__(self, entries: list[tuple[bytes, list[tuple[int, bytes]]]]) -> None:
+        self.entries = entries
+        super().__init__(REUC_EXTENSION, b"")
+
+    @classmethod
+    def from_bytes(cls, data: bytes) -> "ResolveUndoExtension":
+        # TODO: Implement resolve undo parsing
+        return cls([])
+
+    def to_bytes(self) -> bytes:
+        # TODO: Implement resolve undo serialization
+        return b""
+
+
+class UntrackedExtension(IndexExtension):
+    """Untracked cache extension."""
+
+    def __init__(self, data: bytes) -> None:
+        super().__init__(UNTR_EXTENSION, data)
+
+    @classmethod
+    def from_bytes(cls, data: bytes) -> "UntrackedExtension":
+        return cls(data)
+
+
 @dataclass
 class IndexEntry:
     ctime: Union[int, float, tuple[int, int]]
@@ -106,6 +357,8 @@ class IndexEntry:
     gid: int
     size: int
     sha: bytes
+    flags: int = 0
+    extended_flags: int = 0
 
     @classmethod
     def from_serialized(cls, serialized: SerializedIndexEntry) -> "IndexEntry":
@@ -119,9 +372,14 @@ class IndexEntry:
             gid=serialized.gid,
             size=serialized.size,
             sha=serialized.sha,
+            flags=serialized.flags,
+            extended_flags=serialized.extended_flags,
         )
 
     def serialize(self, name: bytes, stage: Stage) -> SerializedIndexEntry:
+        # Clear out any existing stage bits, then set them from the Stage.
+        new_flags = self.flags & ~FLAG_STAGEMASK
+        new_flags |= stage.value << FLAG_STAGESHIFT
         return SerializedIndexEntry(
             name=name,
             ctime=self.ctime,
@@ -133,10 +391,34 @@ class IndexEntry:
             gid=self.gid,
             size=self.size,
             sha=self.sha,
-            flags=stage.value << FLAG_STAGESHIFT,
-            extended_flags=0,
+            flags=new_flags,
+            extended_flags=self.extended_flags,
         )
 
+    def stage(self) -> Stage:
+        return Stage((self.flags & FLAG_STAGEMASK) >> FLAG_STAGESHIFT)
+
+    @property
+    def skip_worktree(self) -> bool:
+        """Return True if the skip-worktree bit is set in extended_flags."""
+        return bool(self.extended_flags & EXTENDED_FLAG_SKIP_WORKTREE)
+
+    def set_skip_worktree(self, skip: bool = True) -> None:
+        """Helper method to set or clear the skip-worktree bit in extended_flags.
+        Also sets FLAG_EXTENDED in self.flags if needed.
+        """
+        if skip:
+            # Turn on the skip-worktree bit
+            self.extended_flags |= EXTENDED_FLAG_SKIP_WORKTREE
+            # Also ensure the main 'extended' bit is set in flags
+            self.flags |= FLAG_EXTENDED
+        else:
+            # Turn off the skip-worktree bit
+            self.extended_flags &= ~EXTENDED_FLAG_SKIP_WORKTREE
+            # Optionally unset the main extended bit if no extended flags remain
+            if self.extended_flags == 0:
+                self.flags &= ~FLAG_EXTENDED
+
 
 class ConflictedIndexEntry:
     """Index entry that represents a conflict."""
@@ -210,11 +492,15 @@ def write_cache_time(f, t) -> None:
     f.write(struct.pack(">LL", *t))
 
 
-def read_cache_entry(f, version: int) -> SerializedIndexEntry:
+def read_cache_entry(
+    f, version: int, previous_path: bytes = b""
+) -> SerializedIndexEntry:
     """Read an entry from a cache file.
 
     Args:
       f: File-like object to read from
+      version: Index version
+      previous_path: Previous entry's path (for version 4 compression)
     """
     beginoffset = f.tell()
     ctime = read_cache_time(f)
@@ -235,11 +521,19 @@ def read_cache_entry(f, version: int) -> SerializedIndexEntry:
         (extended_flags,) = struct.unpack(">H", f.read(2))
     else:
         extended_flags = 0
-    name = f.read(flags & FLAG_NAMEMASK)
+
+    if version >= 4:
+        # Version 4: paths are always compressed (name_len should be 0)
+        name, consumed = _decompress_path_from_stream(f, previous_path)
+    else:
+        # Versions < 4: regular name reading
+        name = f.read(flags & FLAG_NAMEMASK)
+
     # Padding:
     if version < 4:
         real_size = (f.tell() - beginoffset + 8) & ~7
         f.read((beginoffset + real_size) - f.tell())
+
     return SerializedIndexEntry(
         name,
         ctime,
@@ -256,21 +550,35 @@ def read_cache_entry(f, version: int) -> SerializedIndexEntry:
     )
 
 
-def write_cache_entry(f, entry: SerializedIndexEntry, version: int) -> None:
+def write_cache_entry(
+    f, entry: SerializedIndexEntry, version: int, previous_path: bytes = b""
+) -> None:
     """Write an index entry to a file.
 
     Args:
       f: File object
-      entry: IndexEntry to write, tuple with:
+      entry: IndexEntry to write
+      version: Index format version
+      previous_path: Previous entry's path (for version 4 compression)
     """
     beginoffset = f.tell()
     write_cache_time(f, entry.ctime)
     write_cache_time(f, entry.mtime)
-    flags = len(entry.name) | (entry.flags & ~FLAG_NAMEMASK)
+
+    if version >= 4:
+        # Version 4: use compression but set name_len to actual filename length
+        # This matches how C Git implements index v4 flags
+        compressed_path = _compress_path(entry.name, previous_path)
+        flags = len(entry.name) | (entry.flags & ~FLAG_NAMEMASK)
+    else:
+        # Versions < 4: include actual name length
+        flags = len(entry.name) | (entry.flags & ~FLAG_NAMEMASK)
+
     if entry.extended_flags:
         flags |= FLAG_EXTENDED
     if flags & FLAG_EXTENDED and version is not None and version < 3:
         raise AssertionError("unable to use extended flags in version < 3")
+
     f.write(
         struct.pack(
             b">LLLLLL20sH",
@@ -286,8 +594,13 @@ def write_cache_entry(f, entry: SerializedIndexEntry, version: int) -> None:
     )
     if flags & FLAG_EXTENDED:
         f.write(struct.pack(b">H", entry.extended_flags))
-    f.write(entry.name)
-    if version < 4:
+
+    if version >= 4:
+        # Version 4: always write compressed path
+        f.write(compressed_path)
+    else:
+        # Versions < 4: write regular path and padding
+        f.write(entry.name)
         real_size = (f.tell() - beginoffset + 8) & ~7
         f.write(b"\0" * ((beginoffset + real_size) - f.tell()))
 
@@ -299,16 +612,74 @@ class UnsupportedIndexFormat(Exception):
         self.index_format_version = version
 
 
-def read_index(f: BinaryIO) -> Iterator[SerializedIndexEntry]:
-    """Read an index file, yielding the individual entries."""
+def read_index_header(f: BinaryIO) -> tuple[int, int]:
+    """Read an index header from a file.
+
+    Returns:
+      tuple of (version, num_entries)
+    """
     header = f.read(4)
     if header != b"DIRC":
         raise AssertionError(f"Invalid index file header: {header!r}")
     (version, num_entries) = struct.unpack(b">LL", f.read(4 * 2))
-    if version not in (1, 2, 3):
+    if version not in (1, 2, 3, 4):
         raise UnsupportedIndexFormat(version)
+    return version, num_entries
+
+
+def write_index_extension(f: BinaryIO, extension: IndexExtension) -> None:
+    """Write an index extension.
+
+    Args:
+      f: File-like object to write to
+      extension: Extension to write
+    """
+    data = extension.to_bytes()
+    f.write(extension.signature)
+    f.write(struct.pack(">I", len(data)))
+    f.write(data)
+
+
+def read_index(f: BinaryIO) -> Iterator[SerializedIndexEntry]:
+    """Read an index file, yielding the individual entries."""
+    version, num_entries = read_index_header(f)
+    previous_path = b""
     for i in range(num_entries):
-        yield read_cache_entry(f, version)
+        entry = read_cache_entry(f, version, previous_path)
+        previous_path = entry.name
+        yield entry
+
+
+def read_index_dict_with_version(
+    f: BinaryIO,
+) -> tuple[dict[bytes, Union[IndexEntry, ConflictedIndexEntry]], int]:
+    """Read an index file and return it as a dictionary along with the version.
+
+    Returns:
+      tuple of (entries_dict, version)
+    """
+    version, num_entries = read_index_header(f)
+
+    ret: dict[bytes, Union[IndexEntry, ConflictedIndexEntry]] = {}
+    previous_path = b""
+    for i in range(num_entries):
+        entry = read_cache_entry(f, version, previous_path)
+        previous_path = entry.name
+        stage = entry.stage()
+        if stage == Stage.NORMAL:
+            ret[entry.name] = IndexEntry.from_serialized(entry)
+        else:
+            existing = ret.setdefault(entry.name, ConflictedIndexEntry())
+            if isinstance(existing, IndexEntry):
+                raise AssertionError(f"Non-conflicted entry for {entry.name!r} exists")
+            if stage == Stage.MERGE_CONFLICT_ANCESTOR:
+                existing.ancestor = IndexEntry.from_serialized(entry)
+            elif stage == Stage.MERGE_CONFLICT_THIS:
+                existing.this = IndexEntry.from_serialized(entry)
+            elif stage == Stage.MERGE_CONFLICT_OTHER:
+                existing.other = IndexEntry.from_serialized(entry)
+
+    return ret, version
 
 
 def read_index_dict(f) -> dict[bytes, Union[IndexEntry, ConflictedIndexEntry]]:
@@ -348,10 +719,24 @@ def write_index(
     """
     if version is None:
         version = DEFAULT_VERSION
+    # STEP 1: check if any extended_flags are set
+    uses_extended_flags = any(e.extended_flags != 0 for e in entries)
+    if uses_extended_flags and version < 3:
+        # Force or bump the version to 3
+        version = 3
+    # The rest is unchanged, but you might insert a final check:
+    if version < 3:
+        # Double-check no extended flags appear
+        for e in entries:
+            if e.extended_flags != 0:
+                raise AssertionError("Attempt to use extended flags in index < v3")
+    # Proceed with the existing code to write the header and entries.
     f.write(b"DIRC")
     f.write(struct.pack(b">LL", version, len(entries)))
+    previous_path = b""
     for entry in entries:
-        write_cache_entry(f, entry, version)
+        write_cache_entry(f, entry, version=version, previous_path=previous_path)
+        previous_path = entry.name
 
 
 def write_index_dict(
@@ -411,16 +796,25 @@ class Index:
 
     _byname: dict[bytes, Union[IndexEntry, ConflictedIndexEntry]]
 
-    def __init__(self, filename: Union[bytes, str], read=True) -> None:
+    def __init__(
+        self,
+        filename: Union[bytes, str, os.PathLike],
+        read=True,
+        skip_hash: bool = False,
+        version: Optional[int] = None,
+    ) -> None:
         """Create an index object associated with the given filename.
 
         Args:
           filename: Path to the index file
           read: Whether to initialize the index from the given file, should it exist.
+          skip_hash: Whether to skip SHA1 hash when writing (for manyfiles feature)
+          version: Index format version to use (None = auto-detect from file or use default)
         """
-        self._filename = filename
+        self._filename = os.fspath(filename)
         # TODO(jelmer): Store the version returned by read_index
-        self._version = None
+        self._version = version
+        self._skip_hash = skip_hash
         self.clear()
         if read:
             self.read()
@@ -436,10 +830,19 @@ class Index:
         """Write current contents of index to disk."""
         f = GitFile(self._filename, "wb")
         try:
-            f = SHA1Writer(f)
-            write_index_dict(f, self._byname, version=self._version)
-        finally:
+            if self._skip_hash:
+                # When skipHash is enabled, write the index without computing SHA1
+                write_index_dict(f, self._byname, version=self._version)
+                # Write 20 zero bytes instead of SHA1
+                f.write(b"\x00" * 20)
+                f.close()
+            else:
+                f = SHA1Writer(f)
+                write_index_dict(f, self._byname, version=self._version)
+                f.close()
+        except:
             f.close()
+            raise
 
     def read(self) -> None:
         """Read current contents of index from disk."""
@@ -448,10 +851,14 @@ class Index:
         f = GitFile(self._filename, "rb")
         try:
             f = SHA1Reader(f)
-            self.update(read_index_dict(f))
-            # FIXME: Additional data?
-            f.read(os.path.getsize(self._filename) - f.tell() - 20)
-            f.check_sha()
+            entries, version = read_index_dict_with_version(f)
+            self._version = version
+            self.update(entries)
+            # Read any remaining data before the SHA
+            remaining = os.path.getsize(self._filename) - f.tell() - 20
+            if remaining > 0:
+                f.read(remaining)
+            f.check_sha(allow_empty=True)
         finally:
             f.close()
 
@@ -689,15 +1096,17 @@ def index_entry_from_stat(
         mode = cleanup_mode(stat_val.st_mode)
 
     return IndexEntry(
-        stat_val.st_ctime,
-        stat_val.st_mtime,
-        stat_val.st_dev,
-        stat_val.st_ino,
-        mode,
-        stat_val.st_uid,
-        stat_val.st_gid,
-        stat_val.st_size,
-        hex_sha,
+        ctime=stat_val.st_ctime,
+        mtime=stat_val.st_mtime,
+        dev=stat_val.st_dev,
+        ino=stat_val.st_ino,
+        mode=mode,
+        uid=stat_val.st_uid,
+        gid=stat_val.st_gid,
+        size=stat_val.st_size,
+        sha=hex_sha,
+        flags=0,
+        extended_flags=0,
     )
 
 
@@ -712,8 +1121,7 @@ if sys.platform == "win32":
         def __init__(self, errno, msg, filename) -> None:
             super(PermissionError, self).__init__(
                 errno,
-                "Unable to create symlink; "
-                f"do you have developer mode enabled? {msg}",
+                f"Unable to create symlink; do you have developer mode enabled? {msg}",
                 filename,
             )
 
@@ -745,7 +1153,7 @@ def build_file_from_blob(
       target_path: Path to write to
       honor_filemode: An optional flag to honor core.filemode setting in
         config file, default is core.filemode=True, change executable bit
-      symlink: Function to use for creating symlinks
+      symlink_fn: Function to use for creating symlinks
     Returns: stat object for the file
     """
     try:
@@ -963,6 +1371,157 @@ def _has_directory_changed(tree_path: bytes, entry) -> bool:
     return False
 
 
+def update_working_tree(
+    repo,
+    old_tree_id,
+    new_tree_id,
+    honor_filemode=True,
+    validate_path_element=None,
+    symlink_fn=None,
+    force_remove_untracked=False,
+):
+    """Update the working tree and index to match a new tree.
+
+    This function handles:
+    - Adding new files
+    - Updating modified files
+    - Removing deleted files
+    - Cleaning up empty directories
+
+    Args:
+      repo: Repository object
+      old_tree_id: SHA of the tree before the update
+      new_tree_id: SHA of the tree to update to
+      honor_filemode: An optional flag to honor core.filemode setting
+      validate_path_element: Function to validate path elements to check out
+      symlink_fn: Function to use for creating symlinks
+      force_remove_untracked: If True, remove files that exist in working
+        directory but not in target tree, even if old_tree_id is None
+    """
+    import os
+
+    # Set default validate_path_element if not provided
+    if validate_path_element is None:
+        validate_path_element = validate_path_element_default
+
+    # Get the trees
+    old_tree = repo[old_tree_id] if old_tree_id else None
+    repo[new_tree_id]
+
+    # Open the index
+    index = repo.open_index()
+
+    # Track which paths we've dealt with
+    handled_paths = set()
+
+    # Get repo path as string for comparisons
+    repo_path_str = repo.path if isinstance(repo.path, str) else repo.path.decode()
+
+    # First, update/add all files in the new tree
+    for entry in iter_tree_contents(repo.object_store, new_tree_id):
+        handled_paths.add(entry.path)
+
+        # Skip .git directory
+        if entry.path.startswith(b".git"):
+            continue
+
+        # Validate path element
+        if not validate_path(entry.path, validate_path_element):
+            continue
+
+        # Build full path
+        full_path = os.path.join(repo_path_str, entry.path.decode())
+
+        # Get the blob
+        blob = repo.object_store[entry.sha]
+
+        # Ensure parent directory exists
+        parent_dir = os.path.dirname(full_path)
+        if parent_dir and not os.path.exists(parent_dir):
+            os.makedirs(parent_dir)
+
+        # Write the file
+        st = build_file_from_blob(
+            blob,
+            entry.mode,
+            full_path.encode(),
+            honor_filemode=honor_filemode,
+            symlink_fn=symlink_fn,
+        )
+
+        # Update index
+        index[entry.path] = index_entry_from_stat(st, entry.sha)
+
+    # Remove files that existed in old tree but not in new tree
+    if old_tree:
+        for entry in iter_tree_contents(repo.object_store, old_tree_id):
+            if entry.path not in handled_paths:
+                # Skip .git directory
+                if entry.path.startswith(b".git"):
+                    continue
+
+                # File was deleted
+                full_path = os.path.join(repo_path_str, entry.path.decode())
+
+                # Remove from working tree
+                if os.path.exists(full_path):
+                    os.remove(full_path)
+
+                # Remove from index
+                if entry.path in index:
+                    del index[entry.path]
+
+                # Clean up empty directories
+                dir_path = os.path.dirname(full_path)
+                while (
+                    dir_path and dir_path != repo_path_str and os.path.exists(dir_path)
+                ):
+                    try:
+                        if not os.listdir(dir_path):
+                            os.rmdir(dir_path)
+                            dir_path = os.path.dirname(dir_path)
+                        else:
+                            break
+                    except OSError:
+                        break
+
+    # If force_remove_untracked is True, remove any files in working directory
+    # that are not in the target tree (useful for reset --hard)
+    if force_remove_untracked:
+        # Walk through all files in the working directory
+        for root, dirs, files in os.walk(repo_path_str):
+            # Skip .git directory
+            if ".git" in dirs:
+                dirs.remove(".git")
+
+            for file in files:
+                full_path = os.path.join(root, file)
+                # Get relative path from repo root
+                rel_path = os.path.relpath(full_path, repo_path_str)
+                rel_path_bytes = rel_path.encode()
+
+                # If this file is not in the target tree, remove it
+                if rel_path_bytes not in handled_paths:
+                    os.remove(full_path)
+
+                    # Remove from index if present
+                    if rel_path_bytes in index:
+                        del index[rel_path_bytes]
+
+        # Clean up empty directories
+        for root, dirs, files in os.walk(repo_path_str, topdown=False):
+            if ".git" in root:
+                continue
+            if root != repo_path_str and not files and not dirs:
+                try:
+                    os.rmdir(root)
+                except OSError:
+                    pass
+
+    # Write the updated index
+    index.write()
+
+
 def get_unstaged_changes(
     index: Index, root_path: Union[str, bytes], filter_blob_callback=None
 ):
