@@ -26,7 +26,6 @@ import binascii
 import os
 import posixpath
 import stat
-import warnings
 import zlib
 from collections import namedtuple
 from collections.abc import Callable, Iterable, Iterator
@@ -39,6 +38,7 @@ from typing import (
     Union,
 )
 
+from . import replace_me
 from .errors import (
     ChecksumMismatch,
     FileFormatException,
@@ -536,7 +536,7 @@ class ShaFile:
         """Returns the length of the raw string of this object."""
         return sum(map(len, self.as_raw_chunks()))
 
-    def sha(self) -> "HASH":
+    def sha(self) -> Union[FixedSha, "HASH"]:
         """The SHA1 object that is the name of this object."""
         if self._sha is None or self._needs_serialization:
             # this is a local because as_raw_chunks() overwrites self._sha
@@ -892,8 +892,7 @@ class Tag(ShaFile):
     )
     tag_time = serializable_property(
         "tag_time",
-        "The creation timestamp of the tag.  As the number of seconds "
-        "since the epoch",
+        "The creation timestamp of the tag.  As the number of seconds since the epoch",
     )
     tag_timezone = serializable_property(
         "tag_timezone", "The timezone that tag_time is in."
@@ -1321,6 +1320,7 @@ def format_time_entry(person, time, timezone_info):
     )
 
 
+@replace_me(since="0.21.0", remove_in="0.24.0")
 def parse_commit(chunks):
     """Parse a commit object from chunks.
 
@@ -1329,7 +1329,6 @@ def parse_commit(chunks):
     Returns: Tuple of (tree, parents, author_info, commit_info,
         encoding, mergetag, gpgsig, message, extra)
     """
-    warnings.warn("parse_commit will be removed in 0.22", DeprecationWarning)
     parents = []
     extra = []
     tree = None
@@ -1617,13 +1616,9 @@ class Commit(ShaFile):
         doc="Parents of this commit, by their SHA1.",
     )
 
+    @replace_me(since="0.21.0", remove_in="0.24.0")
     def _get_extra(self):
         """Return extra settings of this commit."""
-        warnings.warn(
-            "Commit.extra is deprecated. Use Commit._extra instead.",
-            DeprecationWarning,
-            stacklevel=2,
-        )
         return self._extra
 
     extra = property(
@@ -1644,7 +1639,7 @@ class Commit(ShaFile):
 
     commit_time = serializable_property(
         "commit_time",
-        "The timestamp of the commit. As the number of seconds since the " "epoch.",
+        "The timestamp of the commit. As the number of seconds since the epoch.",
     )
 
     commit_timezone = serializable_property(
