@@ -131,7 +131,7 @@ class GitClientTests(TestCase):
         )
 
     def test_archive_ack(self) -> None:
-        self.rin.write(b"0009NACK\n" b"0000")
+        self.rin.write(b"0009NACK\n0000")
         self.rin.seek(0)
         self.client.archive(b"bla", b"HEAD", None, None)
         self.assertEqual(self.rout.getvalue(), b"0011argument HEAD0000")
@@ -527,6 +527,24 @@ class TestGetTransportAndPath(TestCase):
         self.assertIsInstance(c, LocalGitClient)
         self.assertEqual("foo.bar/baz", path)
 
+    def test_ssh_with_config(self) -> None:
+        # Test that core.sshCommand from config is passed to SSHGitClient
+        from dulwich.config import ConfigDict
+
+        config = ConfigDict()
+        config.set((b"core",), b"sshCommand", b"custom-ssh -o CustomOption=yes")
+
+        c, path = get_transport_and_path(
+            "ssh://git@github.com/user/repo.git", config=config
+        )
+        self.assertIsInstance(c, SSHGitClient)
+        self.assertEqual("custom-ssh -o CustomOption=yes", c.ssh_command)
+
+        # Test rsync-style URL also gets the config
+        c, path = get_transport_and_path("git@github.com:user/repo.git", config=config)
+        self.assertIsInstance(c, SSHGitClient)
+        self.assertEqual("custom-ssh -o CustomOption=yes", c.ssh_command)
+
     @skipIf(sys.platform != "win32", "Behaviour only happens on windows.")
     def test_local_abs_windows_path(self) -> None:
         c, path = get_transport_and_path("C:\\foo.bar\\baz")
@@ -818,6 +836,33 @@ class SSHGitClientTests(TestCase):
         test_client = SSHGitClient("git.samba.org", ssh_command="ssh -o Option1=Value1")
         self.assertEqual(test_client.ssh_command, "ssh -o Option1=Value1")
 
+    def test_ssh_command_config(self) -> None:
+        # Test core.sshCommand config setting
+        from dulwich.config import ConfigDict
+
+        # No config, no environment - should be None
+        self.overrideEnv("GIT_SSH", None)
+        self.overrideEnv("GIT_SSH_COMMAND", None)
+        test_client = SSHGitClient("git.samba.org")
+        self.assertIsNone(test_client.ssh_command)
+
+        # Config with core.sshCommand
+        config = ConfigDict()
+        config.set((b"core",), b"sshCommand", b"ssh -o StrictHostKeyChecking=no")
+        test_client = SSHGitClient("git.samba.org", config=config)
+        self.assertEqual(test_client.ssh_command, "ssh -o StrictHostKeyChecking=no")
+
+        # ssh_command parameter takes precedence over config
+        test_client = SSHGitClient(
+            "git.samba.org", config=config, ssh_command="custom-ssh"
+        )
+        self.assertEqual(test_client.ssh_command, "custom-ssh")
+
+        # Environment variables take precedence over config when no ssh_command parameter
+        self.overrideEnv("GIT_SSH_COMMAND", "/usr/bin/ssh -v")
+        test_client = SSHGitClient("git.samba.org", config=config)
+        self.assertEqual(test_client.ssh_command, "/usr/bin/ssh -v")
+
 
 class ReportStatusParserTests(TestCase):
     def test_invalid_pack(self) -> None:
@@ -1636,8 +1681,7 @@ class PLinkSSHVendorTests(TestCase):
         )
 
         expected_warning = UserWarning(
-            "Invoking PLink with a password exposes the password in the "
-            "process list."
+            "Invoking PLink with a password exposes the password in the process list."
         )
 
         for w in warnings_list:
@@ -1682,8 +1726,7 @@ class PLinkSSHVendorTests(TestCase):
         command = vendor.run_command("host", "git-clone-url", password="12345")
 
         expected_warning = UserWarning(
-            "Invoking PLink with a password exposes the password in the "
-            "process list."
+            "Invoking PLink with a password exposes the password in the process list."
         )
 
         for w in warnings_list:
